#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2021 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype core validation classes.**

This private submodule defines the core low-level class hierarchy driving the
entire :mod:`beartype` validation ecosystem.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                              }....................
# All "FIXME:" comments for this submodule reside in this package's "__init__"
# submodule to improve maintainability and readability here.

# ....................{ IMPORTS                           }....................
from beartype.roar import BeartypeValeSubscriptionException
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.data.utildatadict import merge_mappings_two
from beartype._util.func.utilfuncarg import get_func_args_len_standard
from beartype._util.func.utilfuncscope import CallableScope
from beartype._util.func.utilfunctest import is_func_python
from beartype._util.text.utiltextrepr import represent_object
from typing import Any, Callable

# See the "beartype.cave" submodule for further commentary.
__all__ = ['STAR_IMPORTS_CONSIDERED_HARMFUL']

# ....................{ HINTS                             }....................
SubscriptedIsValidator = Callable[[Any,], bool]
'''
PEP-compliant type hint matching a **validator** (i.e., caller-defined callable
accepting a single arbitrary object and returning either ``True`` if that
object satisfies an arbitrary constraint *or* ``False`` otherwise).

Data validators are suitable for subscripting the :class:`Is` class.
'''

# ....................{ CLASSES ~ subscripted             }....................
class _SubscriptedIs(object):
    '''
    **Beartype validator** (i.e., object encapsulating a caller-defined
    validation callable returning ``True`` when an arbitrary object passed to
    that callable satisfies an arbitrary constraint, suitable for subscripting
    (indexing) `PEP 593`_-compliant :attr:`typing.Annotated` type hints
    enforcing that validation on :mod:`beartype`-decorated callable parameters
    and returns annotated by those hints).

    Caveats
    ----------
    **This low-level class is not intended to be externally instantiated**
    (e.g., by calling the :meth:`__init__` constructor). This class is *only*
    intended to be internally instantiated by subscripting (indexing) the
    higher-level :class:`Is` class factory.

    Attributes
    ----------
    is_valid : Callable[[Any,], bool]
        **Validator** (i.e., caller-defined callable accepting a single
        arbitrary object and returning either ``True`` if that object satisfies
        an arbitrary constraint *or* ``False`` otherwise).
    _is_valid_code : str
        **Validator code** (i.e., Python code snippet validating the
        previously localized parameter or return value against the same
        validation performed by the :meth:`is_valid` function). For efficiency,
        callers validating data through dynamically generated code (e.g., the
        :func:`beartype.beartype` decorator) rather than standard function
        calls (e.g., the private :mod:`beartype._decor._hint._pep._error`
        subpackage) should prefer :attr:`is_valid_code` to :meth:`is_valid`.
        Despite performing the same validation as the :meth:`is_valid`
        callable, this code avoids the additional stack frame imposed by
        calling that callable and thus constitutes an optimization.
    _is_valid_code_locals : CallableScope
        **Validator code local scope** (i.e., dictionary mapping from the name
        to value of each local attribute referenced in :attr:`code`) required
        to dynamically compile this validator code into byte code at runtime.
    _get_repr : Callable[[], str]
        **Representer** (i.e., caller-defined callable accepting *no* arguments
        returning a machine-readable representation of this validator).
        Technically, that representation *could* be passed by the caller rather
        than this callable dynamically generating that representation.
        Pragmatically, generating that representation is sufficiently slow for
        numerous types of validators that deferring their generation until
        required by a call to the :meth:`__repr__` dunder method externally
        called by a call to the :func:`repr` builtin` on this validator is
        effectively mandatory. Data validators whose representations are
        particularly slow to generate include:

        * The :class:`Is` class subscripted by a lambda rather than non-lambda
          function. Generating the representation of that class subscripted by
          a non-lambda function only requires introspecting the name of that
          function and is thus trivially fast. However, lambda functions have
          no names and are thus *only* distinguishable by their source code;
          ergo, generating the representation of that class subscripted by a
          lambda function requires parsing the source code of the file
          declaring that lambda for the exact substring of that code declaring
          that lambda and is thus non-trivially slow.

    See Also
    ----------
    :class:`Is`
        Class docstring for further details.

    .. _PEP 593:
       https://www.python.org/dev/peps/pep-0593
    '''

    # ..................{ CLASS VARIABLES                   }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently called
    # cache dunder methods. Slotting has been shown to reduce read and write
    # costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        'is_valid',
        '_is_valid_code',
        '_is_valid_code_locals',
        '_get_repr',
    )

    # ..................{ INITIALIZERS                      }..................
    def __init__(
        self,
        is_valid: SubscriptedIsValidator,
        is_valid_code: str,
        is_valid_code_locals: CallableScope,
        get_repr: Callable[[], str],
    ) -> None:
        '''
        Initialize this object with the passed validation callable, code, and
        code local scope.

        See the class docstring for usage instructions.

        Parameters
        ----------
        is_valid : Callable[[Any,], bool]
            **Data is_valid** (i.e., caller-defined callable accepting a single
            arbitrary object and returning either ``True`` if that object
            satisfies an arbitrary constraint *or* ``False`` otherwise).
        is_valid_code : str
            **Validator code** (i.e., Python code snippet validating the
            previously localized parameter or return value against the same
            validation performed by the :func:`is_valid` function). This code:

            * *Must* contain one or more ``"{obj}"`` substrings, which external
              code generators (e.g., the :func:`beartype.beartype` decorator)
              will globally replace at evaluation time with the actual test
              subject object to be validated by this code.
            * *May* contain one or more ``"{indent}"`` substrings, which such
              code generators will globally replace at evaluation time with the
              line-oriented indentation required to generate a
              valid Python statement embedding this code. For consistency with
              `PEP 8`_-compliant and well-established Python style guides, any
              additional indentation hard-coded into this code should be
              aligned to **four-space indentation.**
        is_valid_code_locals : Optional[CallableScope]
            **Validator code local scope** (i.e., dictionary mapping from the
            name to value of each local attribute referenced in
            :attr:`is_valid_code` code) required to dynamically compile this
            validator code into byte code at runtime.
        get_repr : Callable[[], str]
            **Representer** (i.e., caller-defined callable accepting *no*
            arguments returning a machine-readable representation of this
            validator). Technically, that representation rather than this
            callable dynamically generating that representation could be passed
            by the caller. Pragmatically, generating that representation is
            sufficiently slow for various types of validators that deferring
            their generation until required by a call to the :meth:`__repr__`
            dunder method externally called by a call to the :func:`repr`
            builtin` passed this validator is effectively mandatory. Data
            validators whose representations are slow to generate include:

            * The :class:`Is` class subscripted by a lambda rather than
              non-lambda function. Generating the representation of that class
              subscripted by a non-lambda function only requires introspecting
              the name of that function and is thus trivially fast. However,
              lambda functions have no names and are thus *only*
              distinguishable by their source code; ergo, generating the
              representation of that class subscripted by a lambda function
              requires parsing the source code of the file declaring that
              lambda for the exact substring of that code declaring that lambda
              and is thus non-trivially slow.

        Raises
        ----------
        BeartypeValeSubscriptionException
            If either:

            * ``is_valid`` is either:

              * *Not* callable.
              * A C-based rather than pure-Python callable.
              * A pure-Python callable accepting two or more arguments.

            * ``is_valid_code`` is either:

              * *Not* a string.
              * A string either:

                * Empty.
                * Non-empty but **invalid** (i.e., *not* containing the test
                  subject substring ``{obj}``).

            * ``is_valid_locals`` is *not* a dictionary.
            * ``get_repr`` is either:

              * *Not* callable.
              * A C-based rather than pure-Python callable.
              * A pure-Python callable accepting one or more arguments.

        .. _PEP 8:
           https://www.python.org/dev/peps/pep-0008
        '''

        # If this validator is uncallable, raise an exception.
        if not callable(is_valid):
            raise BeartypeValeSubscriptionException(
                f'Class "beartype.vale.Is" subscripted argument '
                f'{represent_object(is_valid)} not callable.'
            )
        # Else, this validator is callable.
        #
        # If this validator is C-based, raise an exception.
        elif not is_func_python(is_valid):
            raise BeartypeValeSubscriptionException(
                f'Class "beartype.vale.Is" subscripted callable '
                f'{repr(is_valid)} not pure-Python (e.g., C-based).'
            )
        # Else, this validator is pure-Python.
        #
        # If this validator does *NOT* accept exactly one argument, raise an
        # exception.
        elif get_func_args_len_standard(
            func=is_valid,
            exception_cls=BeartypeValeSubscriptionException,
        ) != 1:
            raise BeartypeValeSubscriptionException(
                f'Class "beartype.vale.Is" subscripted callable '
                f'{repr(is_valid)} positional or keyword argument count '
                f'{get_func_args_len_standard(is_valid)} != 1.'
            )
        # Else, this validator accepts exactly one argument. Since no further
        # validation can be performed on this callable without unsafely calling
        # that callable, we accept this callable as is for now.
        #
        # Note that we *COULD* technically inspect annotations if defined on
        # this callable as well. Since this callable is typically defined as a
        # lambda, annotations are typically *NOT* defined on this callable.

        # If this code is *NOT* a string, raise an exception.
        if not isinstance(is_valid_code, str):
            raise BeartypeValeSubscriptionException(
                f'Data validator code not string:\n'
                f'{represent_object(is_valid_code)}'
            )
        # Else, this code is a string.
        #
        # If this code is the empty string, raise an exception.
        elif not is_valid_code:
            raise BeartypeValeSubscriptionException(
                'Data validator code empty.')
        # Else, this code is a non-empty string.
        #
        # If this code does *NOT* contain the test subject substring
        # "{obj}" and is invalid, raise an exception.
        elif '{obj}' not in is_valid_code:
            raise BeartypeValeSubscriptionException(
                f'Data validator code invalid (i.e., test subject '
                f'substring "{{obj}}" not found):\n{is_valid_code}'
            )
        # Else, this code is hopefully valid.
        #
        # If this code is *NOT* explicitly prefixed by "(" and suffixed by
        # ")", do so to ensure this code remains safely evaluable when
        # embedded in parent expressions.
        elif not (
            is_valid_code[ 0] == '(' and
            is_valid_code[-1] == ')'
        ):
            is_valid_code = f'({is_valid_code})'
        # Else, this code is explicitly prefixed by "(" and suffixed by ")".

        # If this dictionary of code locals is *NOT* a dictionary, raise an
        # exception.
        if not isinstance(is_valid_code_locals, dict):
            raise BeartypeValeSubscriptionException(
                f'Data validator locals '
                f'{represent_object(is_valid_code_locals)} not '
                f'dictionary.'
            )
        # Else, this dictionary of code locals is a dictionary.

        # If this representer is either uncallable, a C-based callable, *OR* a
        # pure-Python callable accepting one or more arguments, raise an
        # exception.
        if get_func_args_len_standard(
            func=get_repr,
            exception_cls=BeartypeValeSubscriptionException,
        ) != 0:
            raise BeartypeValeSubscriptionException(
                f'Representer {repr(get_repr)} positional or keyword argument '
                f'count {get_func_args_len_standard(get_repr)} != 0.'
            )
        # Else, this representer is a pure-Python callable accepting *NO*
        # arguments.

        # Classify this validation function, effectively binding this callable
        # to this object as an object-specific static method.
        self.is_valid = is_valid
        self._is_valid_code = is_valid_code
        self._is_valid_code_locals = is_valid_code_locals
        self._get_repr = get_repr

    # ..................{ DUNDERS ~ operator                }..................
    # Define a domain-specific language (DSL) enabling callers to dynamically
    # combine and Override
    def __and__(self, other: '_SubscriptedIs') -> (
        '_SubscriptedIs'):
        '''
        **Conjunction** (i.e., ``self & other``), synthesizing a new
        :class:`_SubscriptedIs` object whose validator returns ``True`` only
        when the validators of both this *and* the passed
        :class:`_SubscriptedIs` objects all return ``True``.

        Parameters
        ----------
        other : _SubscriptedIs
            Object to conjunctively synthesize with this object.

        Returns
        ----------
        _SubscriptedIs
            New object conjunctively synthesized with this object.

        Raises
        ----------
        BeartypeValeSubscriptionException
            If the passed object is *not* also an instance of the same class.
        '''

        # If the passed object is *NOT* also an instance of this class, raise
        # an exception.
        if not isinstance(other, _SubscriptedIs):
            raise BeartypeValeSubscriptionException(
                f'Subscripted "beartype.vale.Is*" class & operand '
                f'{represent_object(other)} not '
                f'subscripted "beartype.vale.Is*" class.'
            )
        # Else, the passed object is also an instance of this class.

        # Generate code conjunctively performing both validations.
        is_valid_code = f'({self._is_valid_code} and {other._is_valid_code})'

        # Generate locals safely merging the locals required by the code
        # provided by both this and that validator.
        is_valid_code_locals = merge_mappings_two(
            self._is_valid_code_locals, other._is_valid_code_locals)

        # Closures for great justice.
        return _SubscriptedIs(
            is_valid=lambda obj: self.is_valid(obj) and other.is_valid(obj),
            is_valid_code=is_valid_code,
            is_valid_code_locals=is_valid_code_locals,  # type: ignore[arg-type]
            get_repr=lambda: f'{repr(self)} & {repr(other)}',
        )


    def __or__(self, other: '_SubscriptedIs') -> (
        '_SubscriptedIs'):
        '''
        **Disjunction** (i.e., ``self | other``), synthesizing a new
        :class:`_SubscriptedIs` object whose validator returns ``True`` only
        when the validators of either this *or* the passed
        :class:`_SubscriptedIs` objects return ``True``.

        Parameters
        ----------
        other : _SubscriptedIs
            Object to disjunctively synthesize with this object.

        Returns
        ----------
        _SubscriptedIs
            New object disjunctively synthesized with this object.
        '''

        # If the passed object is *NOT* also an instance of this class, raise
        # an exception.
        if not isinstance(other, _SubscriptedIs):
            raise BeartypeValeSubscriptionException(
                f'Subscripted "beartype.vale.Is*" class | operand '
                f'{represent_object(other)} not '
                f'subscripted "beartype.vale.Is*" class.'
            )
        # Else, the passed object is also an instance of this class.

        # Generate code disjunctively performing both validations.
        is_valid_code = f'({self._is_valid_code} or {other._is_valid_code})'

        # Generate locals safely merging the locals required by the code
        # provided by both this and that validator.
        is_valid_code_locals = merge_mappings_two(
            self._is_valid_code_locals, other._is_valid_code_locals)

        # Closures for great justice.
        return _SubscriptedIs(
            is_valid=lambda obj: self.is_valid(obj) or other.is_valid(obj),
            is_valid_code=is_valid_code,
            is_valid_code_locals=is_valid_code_locals,  # type: ignore[arg-type]
            get_repr=lambda: f'{repr(self)} | {repr(other)}',
        )


    #FIXME: Fun optimization: if inverting something that's already been
    #inverted, return the original "_SubscriptedIs" object sans inversion. :p
    def __invert__(self) -> '_SubscriptedIs':
        '''
        **Negation** (i.e., ``~self``), synthesizing a new
        :class:`_SubscriptedIs` object whose validator returns ``True`` only
        when the validators of this :class:`_SubscriptedIs` object returns
        ``False``.

        Returns
        ----------
        _SubscriptedIs
            New object negating this object.
        '''

        # Closures for profound lore.
        return _SubscriptedIs(
            is_valid=lambda obj: not self.is_valid(obj),
            # Inverted validator code, defined as the trivial boolean negation
            # of this validator.
            is_valid_code=f'(not {self._is_valid_code})',
            is_valid_code_locals=self._is_valid_code_locals,
            get_repr=lambda: f'~{repr(self)}',
        )

    # ..................{ DUNDERS ~ str                     }..................
    @callable_cached
    def __repr__(self) -> str:
        '''
        Machine-readable representation of this validator.

        This function is memoized for efficiency.

        Warns
        ----------
        BeartypeValeLambdaWarning
            If this validator is implemented as a pure-Python lambda function
            whose definition is *not* parsable from the script or module
            defining that lambda.
        '''

        # Fight the dark power with... power.
        return self._get_repr()



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2021 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **unmemoized class tester** (i.e., unmemoized and thus efficient
callable testing various properties of arbitrary classes) utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import (
    BeartypeDecorHintPep3119Exception,
    _BeartypeUtilTypeException,
)
from typing import Type

# ....................{ CONSTANTS                          }....................
_MODULE_NAME_BUILTINS = 'builtins'
'''
Fully-qualified name of the module declaring all **builtins** (i.e., objects
defined by the standard :mod:`builtins` module and thus globally available by
default *without* requiring explicit importation).
'''


_MODULE_NAME_BUILTINS_DOTTED = f'{_MODULE_NAME_BUILTINS}.'
'''
Fully-qualified name of the module declaring all builtins followed by a ``.``,
defined purely as a trivial optimization for the frequently accessed
:class:`beartype._decor._cache.cachetype.Beartypistry.__setitem__` dunder
method.
'''

# ....................{ VALIDATORS                        }....................
def die_unless_type(
    # Mandatory parameters.
    cls: object,

    # Optional parameters.
    exception_cls: Type[Exception] = _BeartypeUtilTypeException,
) -> None:
    '''
    Raise an exception unless the passed object is a class.

    Parameters
    ----------
    cls : object
        Object to be validated.
    exception_cls : Type[Exception]
        Type of exception to be raised. Defaults to
        :exc:`_BeartypeUtilTypeException`.

    Raises
    ----------
    exception_cls
        If this object is *not* a class.
    '''

    # If this object is *NOT* a class, raise an exception.
    if not isinstance(cls, type):
        assert isinstance(exception_cls, type), (
            'f{repr(exception_cls)} not exception class.')
        raise exception_cls(f'{repr(cls)} not class.')

# ....................{ VALIDATORS ~ class                }....................
#FIXME: Unit test us up, please.
def die_unless_type_isinstanceable(
    # Mandatory parameters.
    cls: type,

    # Optional parameters.
    cls_label: str = 'Annotated',
    exception_cls: Type[Exception] = BeartypeDecorHintPep3119Exception,
) -> None:
    '''
    Raise an exception unless the passed object is an **isinstanceable class**
    (i.e., class whose metaclass does *not* define an ``__instancecheck__()``
    dunder method that raises an exception).

    Classes that are *not* isinstanceable include most PEP-compliant type
    hints, notably:

    * **Generic aliases** (i.e., subscriptable classes overriding the
      ``__class_getitem__()`` class dunder method standardized by :pep:`560`
      subscripted by an arbitrary object) under Python >= 3.9, whose
      metaclasses define an ``__instancecheck__()`` dunder method to
      unconditionally raise an exception. Generic aliases include:

      * :pep:`484`-compliant **subscripted generics.**
      * :pep:`585`-compliant type hints.

    * User-defined classes whose metaclasses define an ``__instancecheck__()``
      dunder method to unconditionally raise an exception, including:

      * :pep:`544`-compliant protocols *not* decorated by the
        :func:`typing.runtime_checkable` decorator.

    Motivation
    ----------
    When a class whose metaclass defines an ``__instancecheck__()`` dunder
    method is passed as the second parameter to the :func:`isinstance` builtin,
    that builtin defers to that method rather than testing whether the first
    parameter passed to that builtin is an instance of that class. If that
    method raises an exception, that builtin raises the same exception,
    preventing callers from deciding whether arbitrary objects are instances
    of that class. For brevity, we refer to that class as "non-isinstanceable."

    Most classes are isinstanceable, because deciding whether arbitrary objects
    are instances of those classes is a core prerequisite for object-oriented
    programming. Most classes that are also PEP-compliant type hints, however,
    are *not* isinstanceable, because they're *never* intended to be
    instantiated into objects (and typically prohibit instantiation in various
    ways); they're only intended to be referenced as type hints annotating
    callables, an arguably crude form of callable markup.

    :mod:`beartype`-decorated callables typically check the types of arbitrary
    objects at runtime by passing those objects and types as the first and
    second parameters to the :func:`isinstance` builtin. If those types are
    non-isinstanceable, those type-checks will typically raise
    non-human-readable exceptions (e.g., ``"TypeError: isinstance() argument 2
    cannot be a parameterized generic"`` for `PEP 585`_-compliant type hints).
    This is non-ideal both because those exceptions are non-human-readable
    *and* because those exceptions are raised at call rather than decoration
    time, where users expect the :mod:`beartype.beartype` decorator to raise
    exceptions for erroneous type hints.

    Thus the existence of this function, which the :mod:`beartype.beartype`
    decorator calls to validate the usability of type hints that are classes
    *before* checking objects against those classes at call time.

    Parameters
    ----------
    cls : type
        Class to be validated.
    cls_label : str
        Human-readable label prefixing the representation of this class in the
        exception message raised by this function. Defaults to ``"Annotated"``.
    exception_cls : Type[Exception]
        Type of exception to be raised. Defaults to
        :exc:`BeartypeDecorHintPep3119Exception`.

    Raises
    ----------
    BeartypeDecorHintPep3119Exception
        If this hint is *not* an isinstanceable class.
    '''

    # If this hint is *NOT* a class, raise an exception.
    die_unless_type(cls=cls, exception_cls=exception_cls)
    # Else, this hint is a class.

    # If this class is *NOT* isinstanceable, raise an exception. For
    # efficiency, this test is split into two passes (in order of decreasing
    # efficiency):
    #
    # 1. Test whether this class is isinstanceable with the memoized
    #    is_type_isinstanceable() tester. This is crucial, as this test can
    #    *ONLY* be implemented via inefficient EAFP-style exception handling.
    # 2. If that tester reports this class to be non-isinstanceable, raise a
    #    human-readable exception chained onto the non-human-readable exception
    #    raised by explicitly passing that class as the second parameter to the
    #    isinstance() builtin.
    if not is_type_isinstanceable(cls):
        assert isinstance(exception_cls, type), (
            'f{repr(exception_cls)} not exception class.')

        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # CAUTION: Synchronize with the is_type_isinstanceable() tester.
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        try:
            isinstance(None, cls)  # type: ignore[arg-type]
        except Exception as exception:
            #FIXME: Uncomment after we uncover why doing so triggers an
            #infinite circular exception chain when "hint" is a "GenericAlias".
            # # Human-readable exception message to be raised as either...
            # exception_message = (
            #     # If this class is a PEP 544-compliant protocol, a message
            #     # documenting this exact issue and how to resolve it;
            #     (
            #         f'{hint_label} PEP 544 protocol {hint} '
            #         f'uncheckable at runtime (i.e., '
            #         f'not decorated by @typing.runtime_checkable).'
            #     )
            #     if is_hint_pep544_protocol(hint) else
            #     # Else, a fallback message documenting this general issue.
            #     (
            #         f'{hint_label} type {hint} uncheckable at runtime (i.e., '
            #         f'not passable as second parameter to isinstance() '
            #         f'due to raising "{exception}" from metaclass '
            #         f'__instancecheck__() method).'
            #     )
            # )

            # Human-readable exception message to be raised.
            exception_message = (
                f'{cls_label} {repr(cls)} uncheckable at runtime (i.e., '
                f'not passable as second parameter to isinstance() '
                f'due to raising "{exception}" from metaclass '
                f'__instancecheck__() method).'
            )

            # Raise this high-level exception with this human-readable message
            # chained onto this low-level exception with a typically
            # non-human-readable message.
            raise exception_cls(exception_message) from exception

# ....................{ TESTERS ~ builtin                 }....................
def is_type_builtin(cls: type) -> bool:
    '''
    ``True`` only if the passed class is **builtin** (i.e., globally accessible
    C-based type requiring *no* explicit importation).

    This tester is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    cls : type
        Class to be inspected.

    Returns
    ----------
    bool
        ``True`` only if this class is builtin.

    Raises
    ----------
    _BeartypeUtilTypeException
        If this object is *not* a class.
    '''

    # Avoid circular import dependencies.
    from beartype._util.py.utilpymodule import (
        get_object_type_module_name_or_none)

    # If this object is *NOT* a class, raise an exception.
    die_unless_type(cls)
    # Else, this object is a class.

    # If the unqualified basename of this class is "NoneType", this is the
    # class of the "None" singleton. In this case, return false. Unlike all
    # other builtin classes, this class is globally inaccessible despite being
    # declared to be builtin:
    #     >>> import builtins
    #     >>> type(None).__name__
    #     'NoneType'
    #     >>> type(None).__module__
    #     'builtins'
    #     >>> NoneType
    #     NameError: name 'NoneType' is not defined   <---- this is balls
    #
    # This inconsistency almost certainly constitutes a bug in the CPython
    # interpreter, but it seems doubtful anyone else would see it that way and
    # almost certain everyone else would defend this edge case.
    #
    # We're *NOT* dying on that lonely hill. We obey the Law of Guido.
    if cls.__name__ == 'NoneType':
        return False
    # Else, this is *NOT* the class of the "None" singleton.

    # Fully-qualified name of the module defining this class if this class is
    # defined by a module *OR* "None" otherwise (i.e., if this class is
    # dynamically defined in-memory).
    cls_module_name = get_object_type_module_name_or_none(cls)

    # This return true only if this name is that of the "builtins" module
    # declaring all builtin classes.
    return cls_module_name == _MODULE_NAME_BUILTINS


def is_classname_builtin(classname: str) -> bool:
    '''
    ``True`` only if the passed fully-qualified classname is that of a
    **builtin type** (i.e., globally accessible C-based type requiring *no*
    explicit importation).

    This tester is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    classname: str
        Classname to be inspected.

    Returns
    ----------
    bool
        ``True`` only if this classname is that of a builtin type.
    '''

    # Return true only if...
    return (
        # This classname is prefixed by "builtins." *AND*...
        classname.startswith(_MODULE_NAME_BUILTINS_DOTTED) and
        # This classname is *NOT* that of the type of the "None" singleton. See
        # is_type_builtin() for further commentary.
        classname != 'builtins.NoneType'
    )

# ....................{ TESTERS ~ isinstanceable          }....................
def is_type_isinstanceable(cls: object) -> bool:
    '''
    ``True`` only if the passed type cls is an **isinstanceable class** (i.e.,
    class whose metaclass does *not* define an ``__instancecheck__()`` dunder
    method that raises an exception).

    This tester is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator). Although the implementation does *not*
    trivially reduce to an efficient one-liner, the inefficient branch of this
    implementation *only* applies to erroneous edge cases resulting in raised
    exceptions and is thus largely ignorable.

    Caveats
    ----------
    **This tester may return false positives in unlikely edge cases.**
    Internally, this tester tests whether this class is isinstanceable by
    detecting whether passing the ``None`` singleton and this class to the
    :func:`isinstance` builtin raises an exception. If that call raises *no*
    exception, this class is probably but *not* necessarily isinstanceable.
    Since the metaclass of this class could define an ``__instancecheck__()``
    dunder method to conditionally raise exceptions except when passed the
    ``None`` singleton, there exists *no* means of ascertaining whether a class
    is fully isinstanceable in the general case. Since most classes that are
    *not* isinstanceable are unconditionally isinstanceable (i.e., the
    metaclasses of those classes define an ``__instancecheck__()`` dunder
    method to unconditionally raise exceptions), this distinction is generally
    meaningless in the real world. This test thus generally suffices.

    Parameters
    ----------
    cls : object
        Object to be tested.

    Returns
    ----------
    bool
        ``True`` only if this object is an isinstanceable class.

    See Also
    ----------
    :func:`die_unless_type_isinstanceable`
        Further details.
    '''

    # If this object is *NOT* a class, return false.
    if not isinstance(cls, type):
        return False

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Synchronize with die_unless_type_isinstanceable().
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # Attempt to pass this class as the second parameter to the isinstance()
    # builtin to decide whether or not this class is safely usable as a
    # standard class or not.
    #
    # Note that this leverages an EAFP (i.e., "It is easier to ask forgiveness
    # than permission") approach and thus imposes a minor performance penalty,
    # but that there exists *NO* faster alternative applicable to arbitrary
    # user-defined classes, whose metaclasses may define an __instancecheck__()
    # dunder method to raise exceptions and thus prohibit being passed as the
    # second parameter to the isinstance() builtin, the primary means employed
    # by @beartype wrapper functions to check arbitrary types.
    try:
        isinstance(None, cls)

        # If the prior function call raised *NO* exception, this class is
        # probably but *NOT* necessarily isinstanceable.
        return True
    # If the prior function call raised an exception, this class is *NOT*
    # isinstanceable. In this case, return false.
    except:
        return False



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2021 Beartype authors.
# See "LICENSE" for further details.

"""
Project-wide **string munging** (i.e., generic low-level operations
transforming strings into new derivative strings) utilities.

This private submodule is *not* intended for importation by downstream callers.
"""

# ....................{ IMPORTS                           }....................
from beartype.roar._roarwarn import _BeartypeUtilCallableWarning
from beartype._util.utilobject import get_object_scopes_name
from collections.abc import Callable
from re import (
    DOTALL,
    sub as re_sub
)
from string import punctuation
from typing import Any

# ....................{ REPRESENTERS                      }....................
def represent_object(
    # Mandatory parameters.
    obj: object,

    # Optional parameters.
    max_len: int = 76,
    is_strip_from_newline_first: bool = True,
) -> str:
    """
    Pretty-printed quasi-human-readable variant of the string returned by the
    non-pretty-printed machine-readable :meth:`obj.__repr__` dunder method of
    the passed object, truncated to the passed maximum string length.

    Specifically, this function (in order):

    #. Obtains this object's representation by calling ``repr(object)``.
    #. If this representation is neither suffixed by a punctuation character
       (i.e., character in the standard :attr:`string.punctuation` set) *nor*
       representing a byte-string whose representations are prefixed by `b'`
       and suffixed by `'` (e.g., `b'Check, mate.'`), double-quotes this
       representation for disambiguity with preceding characters -- notably,
       sequence indices. Since strings returned by this function commonly
       follow sequence indices in error messages, failing to disambiguate the
       two produces non-human-readable output:

           >>> def wat(mate: typing.List[str]) -> int: return len(mate)
           >>> raise_pep_call_exception(
           ...     func=muh_func, pith_name='mate', pith_value=[7,])
           beartype.roar.BeartypeCallHintPepParamException: @beartyped wat()
           parameter mate=[7] violates PEP type hint typing.List[str], as list
           item 0 value 7 not a str.

       Note the substring "item 0 value 7", which misreads like a blatant bug.
       Double-quoting the "7" suffices to demarcate values from indices.
    #. If this representation exceeds the passed maximum length, replaces the
       suffix of this representation exceeding this length with an ellipses
       (i.e., `"..."` substring).

    Caveats
    ----------
    **This function is unavoidably slow and should thus not be called from
    optimized performance-critical code.** This function internally performs
    mildly expensive operations, including regular expression-based matching
    and substitution. Ideally, this function should *only* be called to create
    user-oriented exception messages where performance is a negligible concern.

    **This function preserves all quote-protected newline characters** (i.e.,
    ``"\\n"``) **in this representation.** Since the :meth:`str.__repr__`
    dunder method implicitly quote-protects all newlines in the original
    string, this function effectively preserves all newlines in strings.

    Parameters
    ----------
    obj : object
        Object to be represented.
    max_len: int, optional
        Maximum length of the string to be returned. Defaults to a standard
        line length of 80 characters minus output indentation of 4 characters.

    Returns
    ----------
    str
        Pretty-printed quasi-human-readable variant of this object's
        non-pretty-printed machine-readable representation.
    """
    assert isinstance(max_len, int), f'{repr(max_len)} not integer.'

    # String describing the passed object. For debuggability, the verbose
    # (albeit less human-readable) output of repr() is preferred to the terse
    # (albeit more human-readable) output of str().
    #
    # Note that this representation quote-protects all newlines in this
    # representation. Ergo, "\n" *MUST* be matched as r"\n" instead.
    obj_repr = repr(obj)

    # If this representation is empty, return empty double-quotes. Although
    # most objects (including outlier singletons like "None" and the empty
    # string) have non-empty representations, caller-defined classes may
    # maliciously override the __repr__() dunder method to return an empty
    # string rather than the representation of an empty string (i.e., '""').
    if not obj_repr:
        return '""'
    # Else, this representation is non-empty.
    #
    # If this representation is neither...
    elif not (
        # Prefixed by punctuation *NOR*...
        obj_repr[0] in punctuation or
        # A byte-string, in which case this representation is instead prefixed
        # by "b'".
        isinstance(obj, bytes)
    ):
    # Then this representation is *NOT* demarcated from preceding characters in
    # the parent string embedding this representation. In this case,
    # double-quote this representation for disambiguity with preceding
    # characters (e.g., sequence indices).
        obj_repr = f'"{obj_repr}"'

    # If this representation exceeds this maximum length...
    if len(obj_repr) > max_len:
        # If this maximum length is long enough to at least allow truncation to
        # ellipses (i.e., a substring of length 3)...
        if max_len > 3:
            #FIXME: Consider compiling and caching this regex into a module
            #dictionary mapping from "max_len" values to compiled regexes.

            # Uncompiled regular expression grouping zero or more leading
            # characters preceding this maximum length *AND* zero or more
            # trailing delimiters.
            #
            # Note that:
            # * This expression conditionally depends on a passed parameter and
            #   thus *CANNOT* be compiled as a module-scoped global.
            # * Efficiency is *NOT* a consideration for this function, which is
            #   only called in the edge-case event of a runtime error.
            PRE_MAX_CHARS_LAST_DELIMITERS_REGEX = (
                # Group anchored at the string start preserving a maximum
                # number (excluding the length required to inject an ellipses)
                # minus one to account for the fact that both bounds of the
                # "{...,...}" range operator are inclusive) of any characters.
                r'^(.{0,'
                f'{max_len - 4}'
                r'})'
                # Ungrouped remainder to be truncated.
                r'.*?'
                # Group anchored at the string end preserving zero or more
                # trailing delimiters.
                r'([\'")}>\]]*)$'
            )

            # Replace the substring of this representation from whichever of the
            # first character following this maximum length or the first newline
            # occurs first to the string end (excluding any optional trailing
            # delimiters) with a single ellipses.
            obj_repr = re_sub(
                PRE_MAX_CHARS_LAST_DELIMITERS_REGEX,
                r'\1...\2',
                obj_repr,
                flags=DOTALL,
            )
        # Else, this maximum length is *NOT* long enough to at least allow
        # truncation to ellipses (i.e., a substring of length 3). In this case,
        # truncate this string to this length *WITHOUT* ellipses.
        else:
            obj_repr = obj_repr[:max_len]

        # print(f'obj repr truncated: {obj_repr}')

    # Return this representation.
    return obj_repr

# ....................{ REPRESENTER ~ callable            }....................
def represent_func(
    # Mandatory parameters.
    func: Callable,

    # Optional parameters.
    warning_cls: type = _BeartypeUtilCallableWarning,
) -> str:
    '''
    Machine-readable representation of the passed callable.

    Caveats
    ----------
    **This function is unavoidably slow and should thus not be called from
    optimized performance-critical code.** This function internally performs
    extremely expensive operations, including abstract syntax tree (AST)-based
    parsing of Python scripts and modules deserialized from disk. Ideally, this
    function should *only* be called to create user-oriented exception messages
    where performance is a negligible concern.

    Parameters
    ----------
    func : Callable
        Callable to be represented.
    warning_cls : type, optional
        Type of warning to be emitted in the event of a non-fatal error.
        Defaults to :class:`_BeartypeUtilCallableWarning`.

    Warns
    ----------
    warning_cls
        If this callable is a pure-Python lambda function whose definition is
        *not* parsable from the script or module defining that lambda.
    '''
    assert callable(func), f'{repr(func)} not callable.'

    # Avoid circular import dependencies.
    from beartype._util.func.utilfunccode import get_func_code_or_none
    from beartype._util.func.utilfunctest import is_func_lambda

    # Return either...
    return (
        # If this callable is a pure-Python lambda function, either:
        # * If this lambda is defined by an on-disk script or module source
        #   file, the exact substring of that file defining this lambda.
        # * Else (e.g., if this lambda is dynamically defined in-memory), a
        #   placeholder string.
        get_func_code_or_none(func=func, warning_cls=warning_cls) or '<lambda>'
        if is_func_lambda(func) else
        # Else, the fully-qualified name of this non-lambda function.
        f'{get_object_scopes_name(func)}()'
    )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2021 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **magic strings** (i.e., globally applicable string constants).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                           }....................
# See the "beartype.cave" submodule for further commentary.
__all__ = ['STAR_IMPORTS_CONSIDERED_HARMFUL']

# ....................{ CODE ~ indent                     }....................
CODE_INDENT_1 = '    '
'''
PEP-agnostic code snippet expanding to a single level of indentation.
'''


CODE_INDENT_2 = CODE_INDENT_1*2
'''
PEP-agnostic code snippet expanding to two levels of indentation.
'''


CODE_INDENT_3 = CODE_INDENT_2 + CODE_INDENT_1
'''
PEP-agnostic code snippet expanding to three levels of indentation.
'''

# ....................{ CODE ~ operator                   }....................
LINE_RSTRIP_INDEX_AND = -len(' and')
'''
Negative index relative to the end of any arbitrary newline-delimited Python
code string suffixed by the boolean operator ``" and"`` required to strip that
suffix from that substring.
'''


LINE_RSTRIP_INDEX_OR = -len(' or')
'''
Negative index relative to the end of any arbitrary newline-delimited Python
code string suffixed by the boolean operator ``" or"`` required to strip that
suffix from that substring.
'''
